from .broker import TaskBroker
from .consumer import TaskConsumer, TaskManager, TaskManagerConfig
from .models import QueuedTask, Task, TaskInfo, TaskPriority, TaskRun, TaskState, task
from .scheduler import TaskScheduler
from .scheduler_crontab import Scheduler, crontab
from .scheduler_every import every

__all__ = [
    "Scheduler",
    "TaskPriority",
    "TaskState",
    "TaskScheduler",
    "task",
    "Task",
    "TaskManagerConfig",
    "TaskManager",
    "TaskConsumer",
    "TaskInfo",
    "TaskRun",
    "QueuedTask",
    "TaskBroker",
    "crontab",
    "every",
]



from __future__ import annotations

from fastapi import FastAPI

from fluid.utils.worker import Workers

from .service import FastapiAppWorkers


def app_workers(app: FastAPI) -> Workers:
    """Get workers from app state or create new workers."""
    if workers := getattr(app.state, "workers", None):
        return workers
    else:
        workers = FastapiAppWorkers.setup(app)
        app.state.workers = workers
        return workers



import asyncio
from typing import Callable


async def wait_for(assertion: Callable[[], bool], timeout: float = 1.0) -> None:
    async with asyncio.timeout(timeout):
        while True:
            if assertion():
                return
            await asyncio.sleep(0)



import json
import os
import uuid
from typing import Any

TRUE_VALUES = frozenset(("yes", "true", "t", "1"))

String = bytes | str


def to_bytes(s: String) -> bytes:
    return s.encode("utf-8") if isinstance(s, str) else s


def to_string(s: String) -> str:
    return s.decode("utf-8") if isinstance(s, bytes) else s


def to_bool(v: str | bool | int | None) -> bool:
    return str(v).lower() in TRUE_VALUES if v else False


def create_uid() -> str:
    return uuid.uuid4().hex


def as_uuid(uid: Any) -> str | None:
    if uid:
        if hasattr(uid, "hex"):
            uid = uid.hex
        try:
            return uuid.UUID(uid).hex
        except ValueError:
            return None
    return None


def nice_env_str(space: int = 4, trim_length: int = 100) -> str:
    lt = max(len(k) for k in os.environ) + space
    values = []
    for key, value in os.environ.items():
        if len(value) > trim_length + 3:
            value = f"{value[:trim_length]}..."
        k = f"{key}:".ljust(lt)
        values.append(f"{k}{value}")
    return "\n".join(values)


def nice_json(data: Any) -> str:
    if not isinstance(data, str):
        return json.dumps(data, indent=4)
    return data


def trim_docstring(docstring: str) -> str:
    """Uniformly trims leading/trailing whitespace from docstrings.
    Based on
    http://www.python.org/peps/pep-0257.html#handling-docstring-indentation
    """
    if not docstring or not docstring.strip():
        return ""
    # Convert tabs to spaces and split into lines
    lines = docstring.expandtabs().splitlines()
    indent = min(len(line) - len(line.lstrip()) for line in lines if line.lstrip())
    trimmed = [lines[0].lstrip()] + [line[indent:].rstrip() for line in lines[1:]]
    return "\n".join(trimmed).strip()



from .cli import DbGroup
from .container import Database
from .crud import CrudDB
from .migration import Migration
from .pagination import Pagination, Search

__all__ = [
    "Database",
    "DbGroup",
    "CrudDB",
    "Migration",
    "Pagination",
    "Search",
]



import os
import signal
from typing import Any, Self

from fastapi import FastAPI

from fluid import settings
from fluid.utils import log
from fluid.utils.worker import Worker, Workers

logger = log.get_logger(__name__)


class FastapiAppWorkers(Workers):

    @classmethod
    def setup(cls, app: FastAPI, **kwargs: Any) -> Self:
        workers = cls(**kwargs)
        app.state.workers = workers
        app.add_event_handler("startup", workers.startup)
        app.add_event_handler("shutdown", workers.shutdown)
        return workers

    def bail_out(self, reason: str, code: int = 1) -> None:
        if settings.ENV != "test":
            logger.warning("shutting down due to %s", reason)
            os.kill(os.getpid(), signal.SIGTERM)

    def get_active_worker(self, *, worker_name: str) -> Worker | None:
        worker = self._workers.get_worker_by_name(worker_name)
        if worker and not worker.is_stopping():
            return worker
        return None



from __future__ import annotations

import asyncio
import enum
import inspect
import logging
import os
import sys
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Coroutine,
    Generic,
    NamedTuple,
    TypeVar,
    overload,
)

from pydantic import BaseModel, Field, field_serializer
from redis.asyncio.lock import Lock
from typing_extensions import Annotated, Doc

from fluid import settings
from fluid.utils import kernel, log
from fluid.utils.data import compact_dict
from fluid.utils.dates import as_utc
from fluid.utils.text import create_uid, trim_docstring

from .errors import TaskDecoratorError, TaskRunError
from .scheduler_crontab import Scheduler

if TYPE_CHECKING:
    from .consumer import TaskManager


TaskExecutor = Callable[["TaskRun"], Coroutine[Any, Any, Any]]
RandomizeType = Callable[[], float | int]
TP = TypeVar("TP", bound=BaseModel)


class EmptyParams(BaseModel):
    pass


class TaskPriority(enum.StrEnum):
    high = enum.auto()
    medium = enum.auto()
    low = enum.auto()


class TaskState(enum.StrEnum):
    init = enum.auto()
    queued = enum.auto()
    running = enum.auto()
    success = enum.auto()
    failure = enum.auto()
    aborted = enum.auto()
    rate_limited = enum.auto()

    @property
    def is_failure(self) -> bool:
        return self is TaskState.failure

    @property
    def is_done(self) -> bool:
        return self in FINISHED_STATES


FINISHED_STATES = frozenset(
    (TaskState.success, TaskState.failure, TaskState.aborted, TaskState.rate_limited)
)


class TaskManagerConfig(BaseModel):
    """Task manager configuration"""

    schedule_tasks: bool = True
    consume_tasks: bool = Field(default=True, description="Consume tasks or sleep")
    max_concurrent_tasks: int = Field(
        default=settings.MAX_CONCURRENT_TASKS,
        description=(
            "The number of coroutine workers consuming tasks. "
            "Each worker consumes one task at a time, therefore, "
            "this number is the maximum number of tasks that can run concurrently."
            "It can be configured via the `FLUID_MAX_CONCURRENT_TASKS` environment "
            "variable, and by default is set to 5."
        ),
    )
    sleep: float = 0.1
    """amount to async sleep after completion of a task"""
    broker_url: str = ""


class TaskInfoBase(BaseModel):
    name: str = Field(description="Task name")
    description: str = Field(description="Task description")
    module: str = Field(description="Task module")
    priority: TaskPriority = Field(description="Task priority")
    schedule: str | None = Field(default=None, description="Task schedule")


class TaskInfoUpdate(BaseModel):
    enabled: bool = Field(default=True, description="Task enabled")
    last_run_end: datetime | None = Field(
        default=None, description="Task last run end as milliseconds since epoch"
    )
    last_run_duration: timedelta | None = Field(
        default=None, description="Task last run duration in milliseconds"
    )
    last_run_state: TaskState | None = Field(
        default=None, description="State of last task run"
    )


class TaskInfo(TaskInfoBase, TaskInfoUpdate):
    pass


class QueuedTask(BaseModel):
    """A task to be queued"""

    run_id: str = Field(description="Task run id")
    task: str = Field(description="Task name")
    params: dict[str, Any] = Field(description="Task parameters")
    priority: TaskPriority | None = Field(default=None, description="Task priority")


class Task(NamedTuple, Generic[TP]):
    """A Task executes any time it is invoked"""

    name: str
    """Task name - unique identifier"""
    executor: TaskExecutor
    """Task executor function"""
    params_model: type[TP]
    """Pydantic model for task parameters"""
    logger: logging.Logger
    """Task logger"""
    module: str = ""
    """Task python module"""
    short_description: str = ""
    """Short task description - one line"""
    description: str = ""
    """Task description - obtained from the executor docstring if not provided"""
    schedule: Scheduler | None = None
    """Task schedule - None means the task is not scheduled"""
    randomize: RandomizeType | None = None
    """Randomize function for task schedule"""
    max_concurrency: int = 0
    """how many tasks can run in each consumer concurrently - 0 means no limit"""
    timeout_seconds: int = 60
    """Task timeout in seconds - how long the task can run before being aborted"""
    priority: TaskPriority = TaskPriority.medium
    """Task priority - high, medium, low"""

    @property
    def cpu_bound(self) -> bool:
        """True if the task is CPU bound"""
        return self.executor is run_in_subprocess

    def info(self, **params: Any) -> TaskInfo:
        params.update(
            name=self.name,
            description=self.description,
            module=self.module,
            priority=self.priority,
            schedule=str(self.schedule) if self.schedule else None,
        )
        return TaskInfo(**compact_dict(params))


class TaskRun(BaseModel, Generic[TP], arbitrary_types_allowed=True):
    """A TaskRun contains all the data generated by a Task run

    This model is never initialized directly, it is created by the TaskManager
    """

    id: str = Field(description="Unique task run id")
    task: Task = Field(description="Task to be executed")
    priority: TaskPriority = Field(description="Task priority")
    params: TP = Field(description="Task parameters")
    state: TaskState = Field(default=TaskState.init, description="Task state")
    task_manager: TaskManager = Field(exclude=True, repr=False)
    queued: datetime | None = None
    start: datetime | None = None
    end: datetime | None = None

    async def execute(self) -> None:
        try:
            self.set_state(TaskState.running)
            await self.task.executor(self)  # type: ignore [arg-type]
        except Exception:
            self.set_state(TaskState.failure)
            raise
        else:
            self.set_state(TaskState.success)

    @field_serializer("task")
    def serialize_task(self, task: Task, _info: Any) -> str:
        return task.name

    @property
    def logger(self) -> logging.Logger:
        return self.task.logger

    @property
    def in_queue(self) -> timedelta | None:
        if self.queued and self.start:
            return self.start - self.queued
        return None

    @property
    def duration(self) -> timedelta | None:
        if self.start and self.end:
            return self.end - self.start
        return None

    @property
    def duration_ms(self) -> float | None:
        duration = self.duration
        if duration is not None:
            return round(1000 * duration.total_seconds(), 2)
        return None

    @property
    def total(self) -> timedelta | None:
        if self.queued and self.end:
            return self.end - self.queued
        return None

    @property
    def name(self) -> str:
        return self.task.name

    @property
    def name_id(self) -> str:
        return f"{self.task.name}.{self.id}"

    @property
    def is_done(self) -> bool:
        return self.state.is_done

    @property
    def is_failure(self) -> bool:
        return self.state.is_failure

    @property
    def deps(self) -> Any:
        return self.task_manager.deps

    def set_state(
        self,
        state: TaskState,
        state_time: datetime | None = None,
    ) -> None:
        if self.state == state:
            return
        state_time = as_utc(state_time)
        match (self.state, state):
            case (TaskState.init, TaskState.queued):
                self.queued = state_time
                self.state = state
                self._dispatch()
            case (TaskState.init, _):
                self.set_state(TaskState.queued, state_time)
                self.set_state(state, state_time)
            case (TaskState.queued, TaskState.running):
                self.start = state_time
                self.state = state
                self._dispatch()
            case (
                TaskState.queued,
                TaskState.success
                | TaskState.aborted
                | TaskState.rate_limited
                | TaskState.failure,
            ):
                self.set_state(TaskState.running, state_time)
                self.set_state(state, state_time)
            case (
                TaskState.running,
                TaskState.success
                | TaskState.aborted
                | TaskState.rate_limited
                | TaskState.failure,
            ):
                self.end = state_time
                self.state = state
                self._dispatch()
            case _:
                raise TaskRunError(f"invalid state transition {self.state} -> {state}")

    def lock(self, timeout: float | None) -> Lock:
        return self.task_manager.broker.lock(self.name, timeout=timeout)

    def _dispatch(self) -> None:
        self.task_manager.dispatcher.dispatch(self.model_copy())  # type: ignore [arg-type]


@dataclass
class TaskRunWaiter:
    task_manager: TaskManager
    uid: str = field(default_factory=create_uid)
    _runs: dict[str, TaskRun] = field(default_factory=dict)

    def event(self, state: TaskState) -> str:
        return f"{state}.{self.uid}"

    def __enter__(self) -> TaskRunWaiter:
        for state in FINISHED_STATES:
            self.task_manager.dispatcher.register_handler(self.event(state), self)
        return self

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        for state in FINISHED_STATES:
            self.task_manager.dispatcher.unregister_handler(self.event(state))

    def __call__(self, task_run: TaskRun) -> None:
        self._runs[task_run.id] = task_run

    async def wait(self, task_run: TaskRun, *, timeout: int = 2) -> TaskRun:
        async with asyncio.timeout(timeout):
            while True:
                if tr := self._runs.get(task_run.id):
                    if tr.is_done:
                        return tr
                await asyncio.sleep(0.01)


@overload
def task(executor: TaskExecutor) -> Task: ...


@overload
def task(
    *,
    name: str | None = None,
    schedule: Scheduler | None = None,
    short_description: str | None = None,
    description: str | None = None,
    randomize: RandomizeType | None = None,
    max_concurrency: int = 0,
    priority: TaskPriority = TaskPriority.medium,
    cpu_bound: bool = False,
    timeout_seconds: Annotated[
        int,
        Doc("Task timeout in seconds - how long the task can run before being aborted"),
    ] = 60,
) -> TaskConstructor: ...


# implementation of the task decorator
def task(executor: TaskExecutor | None = None, **kwargs: Any) -> Task | TaskConstructor:
    """Decorator to create a Task

    This decorator can be used in two ways:

    - As a simple decorator of the executor function
    - As a function with keyword arguments
    """
    if kwargs and executor:
        raise TaskDecoratorError("cannot use positional parameters")
    elif kwargs:
        return TaskConstructor(**kwargs)
    elif not executor:
        raise TaskDecoratorError("this is a decorator cannot be invoked in this way")
    else:
        return TaskConstructor()(executor)


class TaskConstructor:
    def __init__(self, *, cpu_bound: bool = False, **kwargs: Any) -> None:
        self.cpu_bound = cpu_bound
        self.kwargs = kwargs

    def __call__(self, executor: TaskExecutor) -> Task:
        if self.cpu_bound:
            return self.cpu_bound_task(executor)
        else:
            return self.create_task(executor)

    def create_task(
        self,
        executor: TaskExecutor,
        defaults: dict[str, Any] | None = None,
    ) -> Task:
        kwargs: dict[str, Any] = self.kwargs_defaults(executor)
        if defaults:
            kwargs.update(defaults)
        kwargs.update(self.kwargs)
        name = kwargs["name"]
        kwargs.update(
            executor=executor,
            params_model=self.get_params_model(executor),
            logger=log.get_logger(f"task.{name}", prefix=True),
        )
        return Task(**kwargs)

    def cpu_bound_task(self, executor: TaskExecutor) -> Task:
        if is_in_subprocess():
            return self.create_task(executor)
        else:
            return self.create_task(run_in_subprocess, self.kwargs_defaults(executor))

    def kwargs_defaults(self, executor: TaskExecutor) -> dict[str, Any]:
        module = inspect.getmodule(executor)
        description = trim_docstring(inspect.getdoc(executor) or "")
        short_description = description.split("\n")[0].strip()
        return dict(
            name=get_name(executor),
            module=module.__name__ if module else "",
            description=description,
            short_description=short_description,
            executor=executor,
        )

    def get_params_model(self, executor: TaskExecutor) -> type[BaseModel]:
        signature = inspect.signature(executor)
        for p in signature.parameters.values():
            if is_subclass(p.annotation, TaskRun):
                params_model = p.annotation.model_fields["params"]
                a = params_model.annotation
                return a if is_subclass(a, BaseModel) else EmptyParams
        return EmptyParams


def get_name(o: Any) -> str:
    if hasattr(o, "__name__"):
        return str(o.__name__)
    elif hasattr(o, "__class__"):
        return str(o.__class__.__name__)
    else:
        return str(o)


def is_in_subprocess() -> bool:
    return os.getenv("TASK_MANAGER_SPAWN") == "true"


def is_subclass(cls: Any, parent: type) -> bool:
    try:
        return issubclass(cls, parent)
    except TypeError:
        return False


class RemoteLog:
    def __init__(self, out: Any) -> None:
        self.out = out

    def __call__(self, data: bytes) -> None:
        self.out.write(data.decode("utf-8"))


async def run_in_subprocess(ctx: TaskRun[TP]) -> None:
    env = dict(os.environ)
    env["TASK_MANAGER_SPAWN"] = "true"
    result = await kernel.run_python(
        "-W",
        "ignore",
        "-m",
        "fluid.scheduler.cpubound",
        ctx.name,
        ctx.task.module,
        ctx.id,
        ctx.params.model_dump_json(),
        result_callback=RemoteLog(sys.stdout),
        error_callback=RemoteLog(sys.stderr),
        env=env,
        stream_output=True,
        stream_error=True,
    )
    if result:
        raise TaskRunError(result)



import atexit
import logging
import signal
from collections import defaultdict
from contextlib import asynccontextmanager
from dataclasses import dataclass, field
from tempfile import NamedTemporaryFile
from time import monotonic
from typing import AsyncIterator

from fluid import settings

from . import kernel

logger = logging.getLogger(__name__)


class FlamegraphError(RuntimeError):
    pass


@dataclass
class Sampler:
    """
    A simple stack sampler for low-overhead CPU profiling: samples the call
    stack every `interval` seconds and keeps track of counts by frame. Because
    this uses signals, it only works on the main thread.
    """

    interval: float = 0.005
    _started = None
    _stack_counts: defaultdict[str, int] = field(
        default_factory=lambda: defaultdict(int)
    )

    @property
    def started(self) -> float | None:
        return self._started

    def start(self) -> None:
        self.reset()
        signal.signal(signal.SIGVTALRM, self._sample)
        signal.setitimer(signal.ITIMER_VIRTUAL, self.interval)
        atexit.register(self.stop)

    def stop(self) -> None:
        self.reset()
        self._started = None
        signal.setitimer(signal.ITIMER_VIRTUAL, 0)

    def reset(self) -> None:
        self._started = monotonic()
        self._stack_counts = defaultdict(int)

    def stats(self) -> str:
        if self._started is None:
            return ""
        elapsed = monotonic() - self._started
        lines = ["elapsed {}".format(elapsed), "granularity {}".format(self.interval)]
        ordered_stacks = sorted(
            self._stack_counts.items(), key=lambda kv: kv[1], reverse=True
        )
        lines.extend(["{} {}".format(frame, count) for frame, count in ordered_stacks])
        return "\n".join(lines) + "\n"

    @asynccontextmanager
    async def flamegraph_file(self, title: str, stats: str) -> AsyncIterator[bytes]:
        with NamedTemporaryFile(prefix="flamegraph-") as f:
            f.write(stats.encode("utf-8"))
            f.flush()
            result = kernel.CollectBytes()
            error = kernel.CollectBytes()
            code = await kernel.run(
                settings.FLAMEGRAPH_EXECUTABLE,
                f.name,
                "--title",
                title,
                result_callback=result,
                error_callback=error,
            )
            if code:
                raise FlamegraphError(error.data)
            yield result.data

    # INTERNALS

    def _sample(self, signum, frame) -> None:
        if self._started:
            stack = []
            while frame is not None:
                stack.append(self._format_frame(frame))
                frame = frame.f_back
            stack_str = ";".join(reversed(stack))
            self._stack_counts[stack_str] += 1
            signal.setitimer(signal.ITIMER_VIRTUAL, self.interval)

    def _format_frame(self, frame) -> str:
        return "{}({})".format(frame.f_code.co_name, frame.f_globals.get("__name__"))

    def __del__(self):
        self.stop()



import asyncio
from typing import Any

from fluid import settings
from fluid.utils.dates import utcnow
from fluid.utils.worker import WorkerFunction

from .consumer import TaskConsumer
from .scheduler_crontab import CronRun


class TaskScheduler(TaskConsumer):
    """A task manager for scheduling tasks"""

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.add_workers(ScheduleTasks(self))


class ScheduleTasks(WorkerFunction):
    def __init__(
        self,
        task_manager: TaskScheduler,
        heartbeat: float | int = 0.001 * settings.SCHEDULER_HEARTBEAT_MILLIS,
    ) -> None:
        super().__init__(self.tick, heartbeat=heartbeat)
        self.task_manager: TaskScheduler = task_manager
        self.last_run: dict[str, CronRun] = {}

    async def tick(self) -> None:
        if not self.task_manager.config.schedule_tasks:
            return
        now = utcnow()
        periodic_tasks = await self.task_manager.broker.filter_tasks(
            scheduled=True, enabled=True
        )
        for task in periodic_tasks:
            if task.schedule:
                run = task.schedule(now, self.last_run.get(task.name))
                if run:
                    self.last_run[task.name] = run
                    from_now = task.randomize() if task.randomize else 0
                    if from_now:
                        asyncio.get_event_loop().call_later(
                            from_now, self.task_manager.sync_queue, task
                        )
                    else:
                        self.task_manager.sync_queue(task)



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Any, Callable

import click
import uvicorn
from fastapi import FastAPI
from pydanclick import from_pydantic
from pydantic import BaseModel
from rich.console import Console
from rich.table import Table
from uvicorn.importer import import_from_string

from fluid.utils import log as log_
from fluid.utils.lazy import LazyGroup

if TYPE_CHECKING:
    from .consumer import TaskManager
    from .models import TaskRun


TaskManagerApp = FastAPI | Callable[..., Any] | str


class TaskManagerCLI(LazyGroup):
    """CLI for TaskManager

    This class provides a CLI for a TaskManager Application.

    It requires to install the `cli` extra dependencies.
    """

    def __init__(
        self,
        task_manager_app: TaskManagerApp,
        **kwargs: Any,
    ):
        kwargs.setdefault("commands", DEFAULT_COMMANDS)
        super().__init__(**kwargs)
        self.task_manager_app = task_manager_app


def ctx_task_manager_app(ctx: click.Context) -> TaskManagerApp:
    return ctx.parent.command.task_manager_app  # type: ignore


def ctx_app(ctx: click.Context) -> FastAPI:
    app = ctx_task_manager_app(ctx)  # type: ignore
    if isinstance(app, str):
        return import_from_string(app)()
    elif isinstance(app, FastAPI):
        return app
    else:
        return app()


def ctx_task_manager(ctx: click.Context) -> TaskManager:
    return ctx_app(ctx).state.task_manager


class ExecuteTasks(click.Group):
    def list_commands(self, ctx: click.Context) -> list[str]:
        task_manager = ctx_task_manager(ctx)
        return sorted(task_manager.registry)

    def get_command(self, ctx: click.Context, cmd_name: str) -> click.Command | None:
        task_manager = ctx_task_manager(ctx)
        task = task_manager.registry.get(cmd_name)
        if task is None:
            raise click.ClickException(f"Task {cmd_name} not found")

        @click.command(cmd_name, help=task.short_description)
        @click.option("--log", is_flag=True, help="Show logs")
        @from_pydantic(task.params_model)
        def execute_task(log: bool, **kwargs: Any) -> None:
            if log:
                log_.config()
            params = {}
            for value in kwargs.values():
                if isinstance(value, BaseModel):
                    params.update(value.model_dump())
            run = task_manager.execute_sync(cmd_name, **params)
            console = Console()
            console.print(task_run_table(run))

        return execute_task


@click.command()
@click.pass_context
def ls(ctx: click.Context) -> None:
    """List all tasks with their schedules"""
    task_manager = ctx_task_manager(ctx)
    table = asyncio.run(tasks_table(task_manager))
    console = Console()
    console.print(table)


@click.command("serve", short_help="Start app server.")
@click.option(
    "--host",
    "-h",
    default="0.0.0.0",
    help="The interface to bind to",
    show_default=True,
)
@click.option(
    "--port",
    "-p",
    default=8080,
    help="The port to bind to",
    show_default=True,
)
@click.option(
    "--reload",
    is_flag=True,
    default=False,
    help="Enable auto-reload",
    show_default=True,
)
@click.pass_context
def serve(ctx: click.Context, host: str, port: int, reload: bool) -> None:
    """Run the service"""
    task_manager_app = ctx_task_manager_app(ctx)
    uvicorn.run(
        task_manager_app,
        port=port,
        host=host,
        log_level="info",
        reload=reload,
        log_config=log_.config(),
    )


@click.command()
@click.argument("task")
@click.option(
    "--disable",
    is_flag=True,
    default=False,
    help="Disable the task",
    show_default=True,
)
@click.pass_context
def enable(ctx: click.Context, task: str, disable: bool) -> None:
    """Enable or disable a task"""
    task_manager = ctx_task_manager(ctx)
    asyncio.run(enable_task(task_manager, task, not disable))


execute = ExecuteTasks(name="exec", help="Execute a registered task")

DEFAULT_COMMANDS = (ls, execute, serve, enable)


def task_run_table(task_run: TaskRun) -> Table:
    table = Table(title="Task Run", show_header=False)
    color = "red" if task_run.state.is_failure else "green"
    table.add_column("Name", style="cyan")
    table.add_column("Description", style=color)
    table.add_row("name", task_run.task.name)
    table.add_row("description", task_run.task.description)
    table.add_row("run_id", task_run.id)
    table.add_row("state", task_run.state)
    if task_run.start:
        table.add_row("started", task_run.start.isoformat())
    if task_run.end:
        table.add_row("completed", task_run.end.isoformat())
    table.add_row("duration ms", str(task_run.duration_ms))
    return table


async def enable_task(task_manager: TaskManager, task: str, enable: bool) -> None:
    await task_manager.broker.enable_task(task, enable=enable)


async def tasks_table(task_manager: TaskManager) -> Table:
    task_info = await task_manager.broker.get_tasks_info()
    dynamic = {t.name: t for t in task_info}
    table = Table(title="Tasks")
    table.add_column("Name", style="cyan", no_wrap=True)
    table.add_column("Enabled")
    table.add_column("Schedule", style="magenta")
    table.add_column("CPU bound", style="magenta")
    table.add_column("Timeout secs", style="green")
    table.add_column("Priority", style="magenta")
    table.add_column("Description", style="green")
    for name in sorted(task_manager.registry):
        task = task_manager.registry[name]
        table.add_row(
            name,
            ":white_check_mark:" if dynamic[name].enabled else "[red]:x:",
            str(task.schedule),
            ":white_check_mark:" if task.cpu_bound else "[red]:x:",
            str(task.timeout_seconds),
            str(task.priority),
            task.short_description,
        )
    return table



from __future__ import annotations

from typing import TYPE_CHECKING, Callable
from urllib.parse import urlparse

from prometheus_client import Histogram

http_request_duration_seconds = Histogram(
    "http_request_duration_seconds",
    "HTTP request duration seconds",
    ["host", "path", "method", "status"],
)

if TYPE_CHECKING:
    from .http_client import HttpResponse


HttpPathFn = Callable[[str], str]


def monitor_http_call(
    response: HttpResponse,
    duration: float,
    sanitization_fn: HttpPathFn | None = None,
) -> None:
    p = urlparse(response.url)

    path = p.path if not sanitization_fn else sanitization_fn(p.path)
    http_request_duration_seconds.labels(
        p.netloc, path, response.method, str(response.status_code)
    ).observe(duration)



from typing import Annotated, Any, Callable, cast

from fastapi import APIRouter, Depends, FastAPI, HTTPException, Path, Request
from pydantic import BaseModel, Field

from fluid.scheduler import (
    Task,
    TaskInfo,
    TaskManager,
    TaskPriority,
    TaskRun,
)
from fluid.scheduler.errors import UnknownTaskError
from fluid.tools_fastapi import app_workers
from fluid.utils.worker import Worker


def get_task_manger_from_request(request: Request) -> TaskManager:
    return get_task_manger(request.app)


def get_task_manger(app: FastAPI) -> TaskManager:
    return cast(TaskManager, app.state.task_manager)


TaskManagerDep = Annotated[TaskManager, Depends(get_task_manger_from_request)]


class TaskUpdate(BaseModel):
    enabled: bool = Field(description="Task enabled or disabled")


class TaskCreate(BaseModel):
    name: str = Field(description="Task name")
    params: dict[str, Any] = Field(description="Task parameters", default_factory=dict)
    priority: TaskPriority | None = Field(default=None, description="Task priority")


def get_router(task_manager: TaskManager) -> APIRouter:
    router = APIRouter()

    router.add_api_route(
        "/tasks",
        get_tasks,
        methods=["GET"],
        response_model=list[TaskInfo],
        summary="List Tasks",
        description="Retrieve a list of tasks runs",
    )

    router.add_api_route(
        "/tasks-status",
        get_task_status,
        methods=["GET"],
        response_model=dict,
        summary="Task consumer status",
        description="Status of the task consumer",
    )

    router.add_api_route(
        "/tasks/{task_name}",
        get_task,
        methods=["GET"],
        response_model=TaskInfo,
        summary="Get a Task",
        description="Retrieve information about a task",
    )

    router.add_api_route(
        "/tasks/{task_name}",
        patch_task,
        methods=["PATCH"],
        response_model=TaskInfo,
        summary="Update a task",
        description="Update a task configuration and enable/disable it",
    )

    for task in task_manager.registry.values():
        router.add_api_route(
            f"/tasks/{task.name}",
            create_queue_task(task),
            methods=["POST"],
            response_model=TaskRun,
            summary=f"Queue a new {task.name} task",
            description=f"Queue a new {task.name} task to be run",
        )

    return router


def create_queue_task(task: Task) -> Callable:
    TaskParams = task.params_model or BaseModel  # noqa

    async def queue_task(
        task_manager: TaskManagerDep,
        params: TaskParams | None = None,  # type: ignore [valid-type]
    ) -> TaskRun:
        try:
            return await task_manager.queue(
                task.name,
                task.priority,
                params=params.model_dump() if params is not None else {},  # type: ignore
            )
        except UnknownTaskError as exc:
            raise HTTPException(status_code=404, detail="Task not found") from exc

    return queue_task


async def get_tasks(task_manager: TaskManagerDep) -> list[TaskInfo]:
    return await task_manager.broker.get_tasks_info()


async def get_task(
    task_manager: TaskManagerDep,
    task_name: str = Path(title="Task name"),
) -> TaskInfo:
    data = await task_manager.broker.get_tasks_info(task_name)
    if not data:
        raise HTTPException(status_code=404, detail="Task not found")
    return data[0]


async def get_task_status(task_manager: TaskManagerDep) -> dict:
    if isinstance(task_manager, Worker):
        return await task_manager.status()
    return {}


async def patch_task(
    task_manager: TaskManagerDep,
    task_update: TaskUpdate,
    task_name: str = Path(title="Task name"),
) -> TaskInfo:
    try:
        return await task_manager.broker.enable_task(task_name, task_update.enabled)
    except UnknownTaskError as exc:
        raise HTTPException(status_code=404, detail="Task not found") from exc


def setup_fastapi(
    task_manager: TaskManager,
    *,
    app: FastAPI | None = None,
    include_router: bool = True,
    **kwargs: Any,
) -> FastAPI:
    """Setup the FastAPI app and add the task manager to the state"""
    app = app or FastAPI(**kwargs)
    if include_router:
        app.include_router(get_router(task_manager), tags=["Tasks"])
    app.state.task_manager = task_manager
    if isinstance(task_manager, Worker):
        app_workers(app).add_workers(task_manager)
    return app



from functools import partial
from typing import Any

from fastapi import FastAPI

from fluid.utils.backdoor import ConsoleManager


def setup(app: FastAPI, **kwargs: Any) -> None:
    console = ConsoleManager(**kwargs)
    app.add_event_handler("startup", partial(console.on_startup, app))
    app.add_event_handler("shutdown", partial(console.on_cleanup, app))



import os

from .utils.text import to_bool

APP_NAME: str = os.getenv("APP_NAME", "fluid")
ENV: str = os.getenv("PYTHON_ENV", "dev")
LOG_LEVEL = (os.getenv("LOG_LEVEL") or "info").upper()
LOG_HANDLER = os.getenv("LOG_HANDLER", "plain")
PYTHON_LOG_FORMAT = os.getenv(
    "PYTHON_LOG_FORMAT",
    "%(asctime)s %(levelname)s %(name)s %(message)s",
)

# Workers
STOPPING_GRACE_PERIOD: int = int(os.getenv("FLUID_STOPPING_GRACE_PERIOD") or "10")
MAX_CONCURRENT_TASKS: int = int(os.getenv("FLUID_MAX_CONCURRENT_TASKS") or "5")
SCHEDULER_HEARTBEAT_MILLIS: int = int(
    os.getenv("FLUID_SCHEDULER_HEARTBEAT_MILLIS", "100")
)
REDIS_DEFAULT_URL = os.getenv("REDIS_DEFAULT_URL", "redis://localhost:6379")
BROKER_URL: str = os.getenv("FLUID_BROKER_URL", REDIS_DEFAULT_URL)
REDIS_MAX_CONNECTIONS = int(os.getenv("MAX_REDIS_CONNECTIONS", "5"))

# Database
DATABASE = os.getenv(
    "DATABASE", "postgresql+asyncpg://postgres:postgres@localhost:5432/fluid"
)
DATABASE_SCHEMA: str | None = os.getenv("DATABASE_SCHEMA")
DBPOOL_MAX_SIZE: int = int(os.getenv("FLUID_DBPOOL_MAX_SIZE") or "10")
DBPOOL_MAX_OVERFLOW: int = int(os.getenv("FLUID_DBPOOL_MAX_OVERFLOW") or "10")
DBECHO: bool = to_bool(os.getenv("FLUID_DBECHO") or "no")

# HTTP
HTTP_USER_AGENT = os.getenv("HTTP_USER_AGENT", f"python/{APP_NAME}")

# Pagination
DEFAULT_PAGINATION_LIMIT = int(os.getenv("DEFAULT_PAGINATION_LIMIT") or "250")
DEFAULT_PAGINATION_MAX_LIMIT = int(os.getenv("DEFAULT_PAGINATION_MAX_LIMIT") or "500")

# Console backdoor
AIO_BACKDOOR_PORT: int = int(os.environ.get("AIO_BACKDOOR_PORT", "8087"))

# Flamegraph
FLAMEGRAPH_EXECUTABLE: str = os.getenv("FLUID_FLAMEGRAPH_EXECUTABLE", "flamegraph.pl")
STACK_SAMPLER_PERIOD_SECONDS: int = int(
    os.getenv("FLUID_STACK_SAMPLER_PERIOD_SECONDS", "1")
)



from __future__ import annotations

import asyncio
from abc import ABC, abstractmethod
from collections import defaultdict
from typing import Awaitable, Callable, Generic, NamedTuple, Self, TypeVar

MessageType = TypeVar("MessageType")
MessageHandlerType = TypeVar("MessageHandlerType")


class Event(NamedTuple):
    type: str
    tag: str

    @classmethod
    def from_string_or_event(cls, event: str | Self) -> Self:
        if isinstance(event, str):
            return cls.from_string(event)
        return event

    @classmethod
    def from_string(cls, event: str) -> Self:
        bits = event.split(".")
        return cls(bits[0], bits[1] if len(bits) > 1 else "")


class BaseDispatcher(Generic[MessageType, MessageHandlerType], ABC):
    def __init__(self) -> None:
        self._msg_handlers: defaultdict[str, dict[str, MessageHandlerType]] = (
            defaultdict(
                dict,
            )
        )

    def register_handler(
        self,
        event: Event | str,
        handler: MessageHandlerType,
    ) -> MessageHandlerType | None:
        event = Event.from_string_or_event(event)
        previous = self._msg_handlers[event.type].get(event.tag)
        self._msg_handlers[event.type][event.tag] = handler
        return previous

    def unregister_handler(self, event: Event | str) -> MessageHandlerType | None:
        event = Event.from_string_or_event(event)
        return self._msg_handlers[event.type].pop(event.tag, None)

    def get_handlers(
        self,
        message: MessageType,
    ) -> dict[str, MessageHandlerType] | None:
        message_type = str(self.message_type(message))
        return self._msg_handlers.get(message_type)

    @abstractmethod
    def message_type(self, message: MessageType) -> str:
        """return the message type"""


class Dispatcher(BaseDispatcher[MessageType, Callable[[MessageType], None]]):
    """Dispatcher for sync handlers"""

    def dispatch(self, message: MessageType) -> int:
        """dispatch the message"""
        handlers = self.get_handlers(message)
        if handlers:
            for handler in handlers.values():
                handler(message)
        return len(handlers or ())


class AsyncDispatcher(
    BaseDispatcher[MessageType, Callable[[MessageType], Awaitable[None]]],
):
    """Dispatcher for async handlers"""

    async def dispatch(self, message: MessageType) -> int:
        """Dispatch the message and wait for all handlers to complete"""
        handlers = self.get_handlers(message)
        if handlers:
            await asyncio.gather(*[handler(message) for handler in handlers.values()])
        return len(handlers or ())


class SimpleDispatcher(Dispatcher[MessageType]):
    def message_type(self, message: MessageType) -> str:
        return "*"
